/*
 *			GPAC - MPEG-4 Systems C Development Kit
 *
 *			Copyright (c) Cyril Concolato 2005
 *					All rights reserved
 *
 *  This file is part of GPAC / command-line mp21 toolbox
 *
 *  GPAC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GPAC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <gpac/intern/m4_isomedia_dev.h>
#include <stdio.h>

void DumpItemInfo(M21File *m21, u32 itemID, Bool full_dump) 
{
	u32 item_count = M21_GetItemCount(m21);
	if (item_count) {
		u32 i;
		for (i = 0; i < item_count; i++) {
			ItemInfoEntry *item_entry = ChainGetEntry(m21->meta->item_infos->item_infos, i);
			if (item_entry->item_ID == itemID) {
				fprintf(stdout, "Item # %d Info \n\tItemID %d - ItemName %s\n\tContentType %s ", i, itemID, item_entry->item_name, item_entry->content_type);
				if(item_entry->content_encoding){
					fprintf(stdout, "ContentEncoding %s\n", item_entry->content_encoding);
				} else {
					fprintf(stdout, "\n");
				}
			}
		}
	}
}

void DumpFileInfo(M21File *m21)
{
	u32 brand, min;
	char sType[5];
	u32 item_count = M21_GetItemCount(m21);
	
	fprintf(stdout, "* MPEG-21 File Info *\n\t%d Item(s)\n", item_count);
	if (m21->mdat) fprintf(stdout, "\tFile has mdat\n");
	else fprintf(stdout, "\tFile has no mdat\n");

	if (M21_GetFileVersionInfo(m21, &brand, &min, NULL) == M4OK) {
		sType[4] = 0;
		MP4TypeToString(brand, sType);
		fprintf(stdout, "\tFile Brand %s - version %d\n", sType, min);
	}

	if (item_count) {
		u32 i;
		for (i = 0; i < item_count; i++) {
			ItemInfoEntry *item_entry = ChainGetEntry(m21->meta->item_infos->item_infos, i);
			DumpItemInfo(m21, item_entry->item_ID, 1);
		}
	}

}
